import collections
import concurrent.futures
import threading
import time

from . import packlib
from . import poollib
from . import server_socket
from . import thread


class Server:

  def __init__(self, port, name='Server', workers=1, errors=True, **kwargs):
    self.socket = server_socket.ServerSocket(port, name, **kwargs)
    self.loop = thread.Thread(self._loop, name=f'{name}Loop')
    self.methods = {}
    self.jobs = set()
    self.workers = workers
    self.errors = errors
    self.running = False
    self.pool = poollib.ThreadPool(workers, 'default_pool')
    self.postfn_pool = poollib.ThreadPool(1, 'postfn')
    self.postfn_inp = collections.deque()
    self.postfn_out = collections.deque()
    self.metrics = dict(send=0, recv=0, time=time.time())
    self.pools = [self.pool, self.postfn_pool]

  def bind(self, name, workfn, postfn=None, workers=0):
    assert not self.running
    assert name not in self.methods, name
    if workers:
      pool = poollib.ThreadPool(workers, '{name}_pool')
      self.pools.append(pool)
    else:
      pool = self.pool
    active = threading.Semaphore((workers or self.workers) + 1)
    def workfn2(*args):
      active.acquire()
      return workfn(*args)
    self.methods[name] = (workfn2, postfn, pool, active)

  def start(self, block=True):
    assert not self.running
    self.running = True
    self.loop.start()
    if block:
      self.loop.join(timeout=None)

  def close(self, timeout=None):
    assert self.running
    self.socket.shutdown()
    self.running = False
    self.loop.join(timeout)
    self.loop.kill()
    [x.close() for x in self.pools]
    self.socket.close()

  def stats(self):
    now = time.time()
    mets = self.metrics
    self.metrics = dict(send=0, recv=0, time=now)
    dur = now - mets['time']
    stats = {
        'numsend': mets['send'],
        'numrecv': mets['recv'],
        'sendrate': mets['send'] / dur,
        'recvrate': mets['recv'] / dur,
        'jobs': len(self.jobs),
    }
    if any(postfn for _, postfn, _, _ in self.methods.values()):
      stats.update({
          'post_iqueue': len(self.postfn_inp),
          'post_oqueue': len(self.postfn_out),
      })
    return stats

  def __enter__(self):
    self.start(block=False)
    return self

  def __exit__(self, *e):
    self.close()

  def _loop(self):
    while self.running or self.jobs or self.postfn_inp or self.postfn_out:
      while True:  # Loop syntax used to break on error.
        if not self.running:  # Do not accept further requests.
          break
        try:
          addr, data = self.socket.recv(timeout=0.0001)
        except TimeoutError:
          break
        if len(data) < 8:
          self._error(addr, bytes(8), 1, 'Message too short')
          break
        reqnum, data = bytes(data[:8]), data[8:]
        try:
          strlen = int.from_bytes(data[:8], 'little', signed=False)
          name = bytes(data[8: 8 + strlen]).decode('utf-8')
          data = packlib.unpack(data[8 + strlen:])
        except Exception:
          self._error(addr, reqnum, 2, 'Could not decode message')
          break
        if name not in self.methods:
          self._error(addr, reqnum, 3, f'Unknown method {name}')
          break
        self.metrics['recv'] += 1
        workfn, postfn, pool, active = self.methods[name]
        job = pool.submit(workfn, *data)
        job.active = active
        job.postfn = postfn
        job.addr = addr
        job.reqnum = reqnum
        self.jobs.add(job)
        if postfn:
          self.postfn_inp.append(job)
        break  # We do not actually want to loop.
      completed, self.jobs = concurrent.futures.wait(
          self.jobs, 0.0001, concurrent.futures.FIRST_COMPLETED)
      for job in completed:
        try:
          data = job.result()
          if job.postfn:
            data, info = data
          else:
            job.active.release()
          data = packlib.pack(data)
          status = int(0).to_bytes(8, 'little', signed=False)
          self.socket.send(job.addr, job.reqnum, status, *data)
          self.metrics['send'] += 1
        except Exception as e:
          self._error(job.addr, job.reqnum, 4, f'Error in server method: {e}')
      if completed:
        while self.postfn_inp and self.postfn_inp[0].done():
          job = self.postfn_inp.popleft()
          data, info = job.result()
          postjob = self.postfn_pool.submit(job.postfn, info)
          postjob.active = job.active
          self.postfn_out.append(postjob)
      while self.postfn_out and self.postfn_out[0].done():
        postjob = self.postfn_out.popleft()
        postjob.active.release()
        postjob.result()  # Check if there was an error.

  def _error(self, addr, reqnum, status, message):
    status = status.to_bytes(8, 'little', signed=False)
    data = message.encode('utf-8')
    self.socket.send(addr, reqnum, status, data)
    if self.errors:
      raise RuntimeError(message)



import functools

import numpy as np
import pytest
import portal


BATCH_SERVERS = [
    portal.BatchServer,
    functools.partial(portal.BatchServer, process=False),
    functools.partial(portal.BatchServer, shmem=True),
]


class TestBatching:

  @pytest.mark.parametrize('BatchServer', BATCH_SERVERS)
  def test_single_client(self, BatchServer):
    port = portal.free_port()
    server = portal.BatchServer(port)
    def fn(x):
      assert x.shape == (4,)
      return 2 * x
    server.bind('fn', fn, batch=4)
    server.start(block=False)
    client = portal.Client('localhost', port)
    futures = [client.fn(x) for x in range(8)]
    results = [x.result() for x in futures]
    assert (results == 2 * np.arange(8)).all()
    client.close()
    server.close()

  def test_multiple_clients(self):
    port = portal.free_port()
    server = portal.BatchServer(port)
    def fn(x):
      assert x.shape == (4,)
      return 2 * x
    server.bind('fn', fn, batch=4)
    server.start(block=False)
    clients = [portal.Client('localhost', port) for _ in range(8)]
    futures = [x.fn(i) for i, x in enumerate(clients)]
    results = [x.result() for x in futures]
    assert (results == 2 * np.arange(8)).all()
    [x.close() for x in clients]
    server.close()

  def test_multiple_workers(self):
    port = portal.free_port()
    server = portal.BatchServer(port)
    def fn(x):
      assert x.shape == (4,)
      return 2 * x
    server.bind('fn', fn, workers=4, batch=4)
    server.start(block=False)
    clients = [portal.Client('localhost', port) for _ in range(32)]
    futures = [x.fn(i) for i, x in enumerate(clients)]
    results = [x.result() for x in futures]
    assert (results == 2 * np.arange(32)).all()
    [x.close() for x in clients]
    server.close()

  @pytest.mark.parametrize('workers', (1, 10))
  def test_proxy(self, workers):
    inner_port = portal.free_port()
    outer_port = portal.free_port()

    server = portal.Server(inner_port, 'InnerServer')
    server.bind('fn', lambda x: 2 * x)
    server.start(block=False)

    kwargs = dict(name='ProxyClient', maxinflight=4)
    proxy_client = portal.Client('localhost', inner_port, **kwargs)
    proxy_server = portal.BatchServer(
        outer_port, 'ProxyServer', workers=workers)
    proxy_server.bind(
        'fn2', lambda x: proxy_client.fn(x).result(), batch=2)
    proxy_server.start(block=False)

    client = portal.Client('localhost', outer_port, 'OuterClient')
    futures = [client.fn2(x) for x in range(16)]
    results = [future.result() for future in futures]
    assert (results == 2 * np.arange(16)).all()

    proxy_server.close()
    proxy_client.close()
    server.close()
    client.close()

  @pytest.mark.parametrize('data', (
      {'a': np.zeros((3, 2), np.float32), 'b': np.ones((1,), np.uint8)},
      {'a': np.array(12), 'b': [np.ones((1,), np.uint8)]},
      {'a': np.array(12), 'b': ['c', [np.array(1), np.array(2.0)]]},
      [],
      {},
      np.array(12),
      [[{}, []]],
  ))
  def test_tree(self, data):
    print(f'\n{data}')
    port = portal.free_port()
    server = portal.BatchServer(port)
    server.bind('fn', lambda x: x, batch=4)
    server.start(block=False)
    client = portal.Client('localhost', port)
    futures = [client.fn(data) for _ in range(4)]
    results = [x.result() for x in futures]
    for result in results:
      assert portal.tree_equals(result, data)
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(5))
  def test_shape_mismatch(self, repeat):
    port = portal.free_port()
    server = portal.BatchServer(port, errors=False)
    server.bind('fn', lambda x: x, batch=2)
    server.start(block=False)
    client = portal.Client('localhost', port)
    future1 = client.fn({'a': np.array(12)})
    future2 = client.fn(42)
    with pytest.raises(RuntimeError):
      future2.result()
    future3 = client.fn({'a': np.array(42)})
    assert future1.result() == {'a': np.array(12)}
    assert future3.result() == {'a': np.array(42)}
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(3))
  def test_client_drops(self, repeat):
    port = portal.free_port()
    server = portal.BatchServer(port)
    server.bind('fn', lambda x: 2 * x, batch=4)
    server.start(block=False)
    client = portal.Client('localhost', port, name='Client1', autoconn=False)
    client.connect()
    future1 = client.fn(1)
    future2 = client.fn(2)
    client.close()
    client = portal.Client('localhost', port, name='Client2')
    future3 = client.fn(3)
    future4 = client.fn(4)
    with pytest.raises(portal.Disconnected):
      future1.result()
    with pytest.raises(portal.Disconnected):
      future2.result()
    assert future3.result() == 6
    assert future4.result() == 8
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(3))
  def test_server_drops(self, repeat):
    port = portal.free_port()
    server = portal.BatchServer(port)
    server.bind('fn', lambda x: 2 * x, batch=2)
    server.start(block=False)
    client = portal.Client('localhost', port, autoconn=False)
    client.connect()
    future1 = client.fn(1)
    server.close()

    server = portal.BatchServer(port)
    server.bind('fn', lambda x: 2 * x, batch=2)
    server.start(block=False)

    with pytest.raises(portal.Disconnected):
      future1.result()
    client.connect()
    future2 = client.fn(2)
    future3 = client.fn(3)
    assert future2.result() == 4
    assert future3.result() == 6
    server.close()
    client.close()



import os
import threading
import time

import numpy as np
import pytest
import portal


SERVERS = [
    portal.Server,
    portal.BatchServer,
]


class TestServer:

  @pytest.mark.parametrize('Server', SERVERS)
  def test_basic(self, Server):
    port = portal.free_port()
    server = Server(port)
    def fn(x):
      assert x == 42
      return 2 * x
    server.bind('fn', fn)
    server.start(block=False)
    client = portal.Client('localhost', port)
    future = client.call('fn', 42)
    assert future.result() == 84
    client.close()
    server.close()

  @pytest.mark.parametrize('Server', SERVERS)
  def test_server_scope(self, Server):
    def fn(data):
      assert data == {'foo': np.array(1)}
      return {'foo': 2 * data['foo']}
    port = portal.free_port()
    server = Server(port)
    server.bind('fn', fn)
    with server:
      client = portal.Client('localhost', port)
      future = client.fn({'foo': np.array(1)})
      result = future.result()
      assert result['foo'] == 2
      client.close()

  @pytest.mark.parametrize('Server', SERVERS)
  def test_multiple_clients(self, Server):
    port = portal.free_port()
    server = Server(port)
    server.bind('fn', lambda data: data)
    server.start(block=False)
    clients = [portal.Client('localhost', port) for _ in range(10)]
    futures = [client.fn(i) for i, client in enumerate(clients)]
    results = [future.result() for future in futures]
    assert results == list(range(10))
    [x.close() for x in clients]
    server.close()

  @pytest.mark.parametrize('Server', SERVERS)
  def test_multiple_methods(self, Server):
    port = portal.free_port()
    server = Server(port)
    server.bind('add', lambda x, y: x + y)
    server.bind('sub', lambda x, y: x - y)
    with server:
      client = portal.Client('localhost', port)
      assert client.add(3, 5).result() == 8
      assert client.sub(3, 5).result() == -2
      client.close()

  @pytest.mark.parametrize('Server', SERVERS)
  def test_unknown_method(self, Server):
    port = portal.free_port()
    server = Server(port, errors=False)
    server.bind('foo', lambda x: x)
    server.start(block=False)
    client = portal.Client('localhost', port)
    future = client.bar(42)
    try:
      future.result()
      assert False
    except RuntimeError as e:
      assert e.args == ('Unknown method bar',)
      assert True
    client.close()
    server.close()

  @pytest.mark.parametrize('Server', SERVERS)
  def test_server_errors(self, Server):
    port = portal.free_port()

    def server(port):
      server = Server(port, errors=True)
      def fn(x):
        if x == 2:
          raise ValueError(x)
        return x
      server.bind('fn', fn)
      server.start(block=True)

    # For the BatchServer, there are resource leak warnings on crash.
    os.environ['PYTHONWARNINGS'] = 'ignore'
    server = portal.Process(server, port, start=True)
    del os.environ['PYTHONWARNINGS']

    client = portal.Client('localhost', port)
    assert client.fn(1).result() == 1
    assert server.running
    with pytest.raises(RuntimeError):
      client.fn(2).result()

    client.close()
    server.join()
    assert server.exitcode not in (0, None)

  @pytest.mark.parametrize('repeat', range(3))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_postfn_ordered(self, repeat, Server):
    completed = []
    logged = []

    def workfn(x):
      if x in (0, 2, 6):
        time.sleep(0.1)
      completed.append(x)
      return x, x

    def postfn(x):
      logged.append(x)

    port = portal.free_port()
    server = Server(port, workers=4)
    server.bind('fn', workfn, postfn)
    server.start(block=False)
    client = portal.Client('localhost', port)
    futures = [client.fn(x) for x in range(10)]
    results = [x.result() for x in futures]
    server.close()
    client.close()
    assert results == list(range(10))
    assert completed != list(range(10))
    assert logged == list(range(10))

  @pytest.mark.parametrize('repeat', range(5))
  @pytest.mark.parametrize('Server', SERVERS)
  @pytest.mark.parametrize('workers', (1, 4))
  def test_postfn_no_backlog(self, repeat, Server, workers):
    port = portal.free_port()
    lock = threading.Lock()
    work_calls = [0]
    done_calls = [0]

    def workfn(x):
      with lock:
        work_calls[0] += 1
        print(work_calls[0], done_calls[0])
        assert work_calls[0] <= done_calls[0] + workers + 1
      return x, x

    def postfn(x):
      with lock:
        done_calls[0] += 1
      time.sleep(0.01)

    server = Server(port, workers=workers)
    server.bind('fn', workfn, postfn)
    server.start(block=False)
    client = portal.Client('localhost', port)
    futures = [client.fn(i) for i in range(20)]
    [future.result() for future in futures]
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(3))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_shared_pool(self, repeat, Server):

    def slow(x):
      time.sleep(0.2)
      return x

    def fast(x):
      return x

    port = portal.free_port()
    server = Server(port, workers=1)
    server.bind('slow', slow)
    server.bind('fast', fast)
    server.start(block=False)
    client = portal.Client('localhost', port)
    slow_future = client.slow(0)
    fast_future = client.fast(0)
    assert not slow_future.done()
    # The slow request is processed first, so this will wait until both
    # requests are done.
    fast_future.result()
    assert slow_future.wait(0.01)
    server.close()
    client.close()

  @pytest.mark.parametrize('repeat', range(3))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_separate_pools(self, repeat, Server):

    def slow(x):
      time.sleep(0.1)
      return x

    def fast(x):
      return x

    port = portal.free_port()
    server = Server(port)
    server.bind('slow', slow, workers=1)
    server.bind('fast', fast, workers=1)
    server.start(block=False)
    client = portal.Client('localhost', port)
    slow_future = client.slow(0)
    fast_future = client.fast(0)
    # Both requests are processed in parallel, so the fast request returns
    # before the slow request is done.
    fast_future.result()
    assert not slow_future.wait(0.01)
    server.close()
    client.close()

  @pytest.mark.parametrize('Server', SERVERS)
  @pytest.mark.parametrize('workers', (1, 10))
  def test_proxy(self, Server, workers):
    inner_port = portal.free_port()
    outer_port = portal.free_port()

    server = Server(inner_port, 'InnerServer')
    server.bind('fn', lambda x: 2 * x)
    server.start(block=False)

    kwargs = dict(name='ProxyClient', maxinflight=4)
    proxy_client = portal.Client('localhost', inner_port, **kwargs)
    proxy_server = Server(outer_port, 'ProxyServer', workers=workers)
    proxy_server.bind('fn2', lambda x: proxy_client.fn(x).result())
    proxy_server.start(block=False)

    client = portal.Client('localhost', outer_port, 'OuterClient')
    futures = [client.fn2(x) for x in range(20)]
    results = [future.result() for future in futures]
    assert results == list(range(0, 40, 2))

    proxy_server.close()
    proxy_client.close()
    server.close()
    client.close()

  @pytest.mark.parametrize('Server', SERVERS)
  def test_sharray(self, Server):
    done = portal.context.mp.Event()

    def server(port, done):
      server = Server(port)
      def fn(data):
        assert data.array.shape == (3, 2)
        assert data.array.dtype == np.float32
        return data
      server.bind('fn', fn)
      server.start(block=False)
      done.wait()
      server.close()

    def client(port):
      data = portal.SharedArray((3, 2), np.float32)
      data.array[:] = np.arange(6, dtype=np.float32).reshape(3, 2)
      client = portal.Client('localhost', port)
      result = client.call('fn', data).result()
      assert result.name == data.name
      assert result, data
      client.close()

    port = portal.free_port()
    client = portal.Process(client, port, start=True)
    server = portal.Process(server, port, done, start=True)
    client.join()
    done.set()
    server.join()

  @pytest.mark.parametrize('repeat', range(3))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_client_drops(self, repeat, Server):
    barrier = threading.Barrier(2)

    def fn(x):
      if x == 1:
        barrier.wait()
        time.sleep(0.2)
      return x

    port = portal.free_port()
    server = Server(port)
    server.bind('fn', fn)
    server.start(block=False)

    client = portal.Client('localhost', port)
    client.fn(1)
    barrier.wait()
    client.close()
    stats = server.stats()
    assert stats['numrecv'] == 1
    assert stats['numsend'] == 0

    client = portal.Client('localhost', port)
    assert client.fn(2).result() == 2
    client.close()
    server.close()



import os
import time

import pytest
import portal


class TestProcess:

  def test_exitcode(self):
    worker = portal.Process(lambda: None, start=True)
    worker.join()
    assert worker.exitcode == 0
    worker = portal.Process(lambda: 42, start=True)
    worker.join()
    assert worker.exitcode == 42

  def test_error(self):
    def fn():
      raise KeyError('foo')
    worker = portal.Process(fn, start=True)
    worker.join()
    assert not worker.running
    assert worker.exitcode == 1

  def test_error_with_children(self):
    def hang():
      while True:
        time.sleep(0.1)
    def fn():
      portal.Process(hang, start=True)
      portal.Thread(hang, start=True)
      time.sleep(0.1)
      raise KeyError('foo')
    worker = portal.Process(fn, start=True)
    worker.join()
    assert not worker.running
    assert worker.exitcode == 1

  def test_kill(self):
    def fn():
      while True:
        time.sleep(0.1)
    worker = portal.Process(fn, start=True)
    worker.kill()
    assert not worker.running
    assert worker.exitcode < 0

  @pytest.mark.parametrize('repeat', range(5))
  def test_kill_with_subproc(self, repeat):
    ready = portal.context.mp.Barrier(3)
    queue = portal.context.mp.Queue()

    def outer(ready, queue):
      queue.put(os.getpid())
      portal.Process(inner, ready, queue, start=True)
      ready.wait()
      while True:
        time.sleep(0.1)

    def inner(ready, queue):
      queue.put(os.getpid())
      ready.wait()
      while True:
        time.sleep(0.1)

    worker = portal.Process(outer, ready, queue, start=True)
    ready.wait()
    worker.kill()
    assert not worker.running
    assert worker.exitcode < 0
    assert not portal.proc_alive(queue.get())
    assert not portal.proc_alive(queue.get())

  @pytest.mark.parametrize('repeat', range(5))
  def test_kill_with_subthread(self, repeat):
    ready = portal.context.mp.Event()
    def outer(ready):
      portal.Thread(inner, ready, start=True)
      while True:
        time.sleep(0.1)
    def inner(ready):
      ready.set()
      while True:
        time.sleep(0.1)
    worker = portal.Process(outer, ready, start=True)
    ready.wait()
    worker.kill()
    assert not worker.running
    assert worker.exitcode < 0

  def test_initfn(self):
    def init():
      portal.foo = 42
    portal.initfn(init)
    ready = portal.context.mp.Event()
    assert portal.foo == 42
    def outer(ready):
      assert portal.foo == 42
      portal.Process(inner, ready, start=True).join()
    def inner(ready):
      assert portal.foo == 42
      ready.set()
    portal.Process(outer, ready, start=True).join()
    ready.wait()
    assert ready.is_set()
    portal.context.initfns.clear()



import collections
import functools
import itertools
import threading
import time
import weakref

from . import client_socket
from . import packlib


class Client:

  def __init__(
      self, host, port=None, name='Client', maxinflight=16, **kwargs):
    assert 1 <= maxinflight, maxinflight
    self.maxinflight = maxinflight
    self.reqnum = iter(itertools.count(0))
    self.futures = {}
    self.errors = collections.deque()
    self.sendrate = [0, time.time()]
    self.recvrate = [0, time.time()]
    self.waitmean = [0, 0]
    self.cond = threading.Condition()
    self.lock = threading.Lock()
    # Socket is created after the above attributes because the callbacks access
    # some of the attributes.
    self.socket = client_socket.ClientSocket(
        host, port, name, start=False, **kwargs)
    self.socket.callbacks_recv.append(self._recv)
    self.socket.callbacks_disc.append(self._disc)
    self.socket.callbacks_conn.append(self._conn)
    self.socket.start()

  @property
  def connected(self):
    return self.socket.connected

  def __getattr__(self, name):
    if name.startswith('_'):
      raise AttributeError(name)
    try:
      return functools.partial(self.call, name)
    except AttributeError:
      raise ValueError(name)

  def stats(self):
    now = time.time()
    stats = {
        'inflight': len(self.futures),
        'numsend': self.sendrate[0],
        'numrecv': self.recvrate[0],
        'sendrate': self.sendrate[0] / (now - self.sendrate[1]),
        'recvrate': self.recvrate[0] / (now - self.recvrate[1]),
        'waitmean': self.waitmean[0] and (self.waitmean[1] / self.waitmean[0]),
    }
    self.sendrate = [0, now]
    self.recvrate = [0, now]
    self.waitmean = [0, 0]
    return stats

  def connect(self, timeout=None):
    return self.socket.connect(timeout)

  def call(self, method, *data):
    reqnum = next(self.reqnum).to_bytes(8, 'little', signed=False)
    start = time.time()
    while len(self.futures) >= self.maxinflight:
      with self.cond: self.cond.wait(timeout=0.2)
      try:
        self.socket.require_connection(timeout=0)
      except TimeoutError:
        pass
    with self.lock:
      self.waitmean[1] += time.time() - start
      self.waitmean[0] += 1
      self.sendrate[0] += 1
    if self.errors:  # Raise errors of dropped futures.
      raise self.errors.popleft()
    name = method.encode('utf-8')
    strlen = len(name).to_bytes(8, 'little', signed=False)
    sendargs = (reqnum, strlen, name, *packlib.pack(data))
    rai = [False]
    future = Future(rai)
    future.sendargs = sendargs
    self.futures[reqnum] = future
    # Store future before sending request because the response may come fast
    # and the response handler runs in the socket's background thread.
    try:
      self.socket.send(*sendargs)
    except client_socket.Disconnected:
      future = self.futures.pop(reqnum)
      future.rai[0] = True
      raise
    return future

  def close(self, timeout=None):
    for future in self.futures.values():
      self._seterr(future, client_socket.Disconnected)
    self.futures.clear()
    self.socket.close(timeout)

  def _recv(self, data):
    assert len(data) >= 16, 'Unexpectedly short response'
    reqnum = bytes(data[:8])
    status = int.from_bytes(data[8:16], 'little', signed=False)
    future = self.futures.pop(reqnum, None)
    if not future:
      existing = sorted(self.futures.keys())
      print(f'Unexpected request number: {reqnum}', existing)
    elif status == 0:
      data = packlib.unpack(data[16:])
      future.set_result(data)
      with self.cond: self.cond.notify_all()
    else:
      message = bytes(data[16:]).decode('utf-8')
      self._seterr(future, RuntimeError(message))
      with self.cond: self.cond.notify_all()
    try:
      self.socket.recv()
    except AssertionError:
      pass  # Socket is already closed.

  def _disc(self):
    if self.socket.options.autoconn:
      for future in list(self.futures.values()):
        future.resend = True
    else:
      for future in list(self.futures.values()):
        self._seterr(future, client_socket.Disconnected)
      self.futures.clear()

  def _conn(self):
    if self.socket.options.autoconn:
      for future in list(self.futures.values()):
        if getattr(future, 'resend', False):
          self.socket.send(*future.sendargs)

  def _seterr(self, future, e):
    future.set_error(e)
    rai = future.rai
    weakref.finalize(future, lambda: (
        None if rai[0] else self.errors.append(e)))


class Future:

  def __init__(self, rai):
    assert rai == [False]
    self.rai = rai
    self.con = threading.Condition()
    self.don = False
    self.res = None
    self.err = None

  def __repr__(self):
    if not self.done:
      return 'Future(done=False)'
    elif self.err:
      return f"Future(done=True, error='{self.err}', raised={self.rai[0]})"
    else:
      return 'Future(done=True)'

  def wait(self, timeout=None):
    if self.don:
      return self.don
    with self.con: return self.con.wait(timeout)

  def done(self):
    return self.don

  def result(self, timeout=None):
    if not self.wait(timeout):
      raise TimeoutError
    assert self.don
    if self.err is None:
      return self.res
    if not self.rai[0]:
      self.rai[0] = True
      raise self.err

  def set_result(self, result):
    assert not self.don
    self.don = True
    self.res = result
    with self.con: self.con.notify_all()

  def set_error(self, e):
    assert not self.don
    self.don = True
    self.err = e
    with self.con: self.con.notify_all()



import collections
import dataclasses
import queue
import select
import socket
import sys
import threading
import time

from . import buffers
from . import contextlib
from . import thread


class Disconnected(Exception):
  pass


@dataclasses.dataclass
class Options:

  ipv6: bool = False
  autoconn: bool = True
  max_msg_size: int = 4 * 1024 ** 3
  max_recv_queue: int = 128
  max_send_queue: int = 128
  keepalive_after: float = 10
  keepalive_every: float = 10
  keepalive_fails: int = 10
  logging: bool = True
  logging_color: str = 'yellow'


class ClientSocket:

  def __init__(self, host, port=None, name='Client', start=True, **kwargs):
    assert (port or ':' in host) and '://' not in host, host
    if port is None:
      host, port = host.rsplit(':', 1)
      port = int(port)
      assert host, host
    self.addr = (host, port)
    self.name = name
    self.options = Options(**{**contextlib.context.clientkw, **kwargs})

    self.callbacks_recv = []
    self.callbacks_conn = []
    self.callbacks_disc = []

    self.isconn = threading.Event()
    self.wantconn = threading.Event()
    self.sendq = collections.deque()
    self.recvq = queue.Queue()

    self.running = True
    self.thread = thread.Thread(self._loop, name=f'{name}Loop')
    start and self.thread.start()

  def start(self):
    self.thread.start()

  @property
  def connected(self):
    return self.isconn.is_set()

  def connect(self, timeout=None):
    if not self.options.autoconn:
      self.wantconn.set()
    return self.isconn.wait(timeout)

  def send(self, *data, timeout=None):
    assert self.running
    if len(self.sendq) > self.options.max_send_queue:
      raise RuntimeError('Too many outgoing messages enqueued')
    self.require_connection(timeout)
    maxsize = self.options.max_msg_size
    self.sendq.append(buffers.SendBuffer(*data, maxsize=maxsize))

  def recv(self, timeout=None):
    assert self.running
    try:
      if timeout is not None and timeout <= 0.2:
        return self.recvq.get(block=(timeout != 0), timeout=timeout)
      start = time.time()
      while True:
        try:
          return self.recvq.get(timeout=min(timeout, 0.2) if timeout else 0.2)
        except queue.Empty:
          timeout = timeout and max(0, timeout - (time.time() - start))
          self.require_connection(timeout)
          if timeout == 0:
            raise
    except queue.Empty:
      raise TimeoutError

  def close(self, timeout=None):
    self.running = False
    self.thread.join(timeout)
    self.thread.kill()

  def require_connection(self, timeout):
    if self.connected:
      return
    if not self.options.autoconn:
      raise Disconnected
    if timeout == 0 or not self.isconn.wait(timeout):
      raise TimeoutError

  def _loop(self):
    recvbuf = buffers.RecvBuffer(maxsize=self.options.max_msg_size)
    sock = None
    poll = select.poll()
    isconn = False  # Local mirror of self.isconn without the lock.

    while self.running or (self.sendq and isconn):

      if not isconn:
        if not self.options.autoconn and not self.wantconn.wait(timeout=0.2):
          continue
        sock = self._connect()
        if not sock:
          break
        poll.register(sock, select.POLLIN | select.POLLOUT)
        self.isconn.set()
        isconn = True
        if not self.options.autoconn:
          self.wantconn.clear()
        [x() for x in self.callbacks_conn]

      try:

        # TODO: According to the py-spy profiler, the GIL is held during
        # polling. Is there a way to avoid that?
        pairs = poll.poll(0.2)
        if not pairs:
          continue
        _, mask = pairs[0]

        if mask & select.POLLIN:
          try:
            recvbuf.recv(sock)
            if recvbuf.done():
              if self.recvq.qsize() > self.options.max_recv_queue:
                raise RuntimeError('Too many incoming messages enqueued')
              msg = recvbuf.result()
              self.recvq.put(msg)
              [x(msg) for x in self.callbacks_recv]
              recvbuf = buffers.RecvBuffer(maxsize=self.options.max_msg_size)
          except BlockingIOError:
            pass

        if self.sendq and mask & select.POLLOUT:
          try:
            self.sendq[0].send(sock)
            if self.sendq[0].done():
              self.sendq.popleft()
          except BlockingIOError:
            pass

      except OSError as e:
        detail = f'{type(e).__name__}'
        detail = f'{detail}: {e}' if str(e) else detail
        self._log(f'Connection to server lost ({detail})')
        self.isconn.clear()
        isconn = False
        poll.unregister(sock)
        sock.close()
        # Clear message queue on disconnect. There is no meaningful concept of
        # sucessful delivery of a message at this level. For example, the
        # server could receive the message but then go down immediately after,
        # without doing anything meaningful with the message. Resending can be
        # done based on response messages at a higher level.
        self.sendq.clear()
        recvbuf = buffers.RecvBuffer(maxsize=self.options.max_msg_size)
        [x() for x in self.callbacks_disc]
        continue

    if sock:
      sock.close()

  def _connect(self):
    host, port = self.addr
    self._log(f'Connecting to {host}:{port}')
    once = True
    while self.running:
      sock, addr = self._create()
      error = None
      try:
        # We need to resolve the address regularly.
        if contextlib.context.resolver:
          addr = contextlib.context.resolver(addr)
        sock.settimeout(10)
        sock.connect(addr)
        sock.settimeout(0)
        self._log('Connection established')
        return sock
      except TimeoutError as e:
        error = e
      except ConnectionError as e:
        error = e
      except socket.gaierror as e:
        error = e
      time.sleep(0.1)
      if once:
        self._log(f'Still trying to connect... ({error})')
        once = False
      sock.close()
    return None

  def _create(self):
    if self.options.ipv6:
      sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
      addr = (*self.addr, 0, 0)
    else:
      sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      addr = self.addr
    # sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    after = self.options.keepalive_after
    every = self.options.keepalive_every
    fails = self.options.keepalive_fails
    if sys.platform == 'linux':
      sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
      sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after)
      sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, every)
      sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, fails)
      sock.setsockopt(
          socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT,
          1000 * (after + every * fails))
    if sys.platform == 'darwin':
      sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
      sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPALIVE, every)
    if sys.platform == 'win32':
      sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, after * 1000, every * 1000))
    return sock, addr

  def _log(self, *args):
    if not self.options.logging:
      return
    contextlib.context.print(
        self.name, *args, color=self.options.logging_color)



import math
import weakref
from multiprocessing import shared_memory

import numpy as np


class SharedArray:

  def __init__(self, shape, dtype, name=None):
    if name:
      self.shm = shared_memory.SharedMemory(name=name)
    else:
      size = math.prod(shape) * np.dtype(dtype).itemsize
      self.shm = shared_memory.SharedMemory(create=True, size=size)
    self.arr = np.ndarray(shape, dtype, self.shm.buf)
    # This unlinks the shared memory buffer, but it will survive until the last
    # process closes up their file pointer. This could cause a problem if a
    # SharedArray is serialize and the Python object goes out of scope before
    # trying to deserialize it (in the same or a different process).
    weakref.finalize(self.arr, self.close)

  @property
  def name(self):
    return self.shm.name

  @property
  def array(self):
    return self.arr

  def result(self):
    return self.arr

  def close(self):
    self.arr = None
    try:
      self.shm.unlink()
    except FileNotFoundError:
      pass

  def __getstate__(self):
    return (self.arr.shape, self.arr.dtype.str, self.shm.name)

  def __setstate__(self, args):
    self.__init__(*args)



import pytest
import portal
import numpy as np


class TestPack:

  @pytest.mark.parametrize('value', [
      np.asarray(42),
      bytes(7),
      {'foo': np.zeros((2, 4), np.float64), 'bar': np.asarray(1)},
      {'foo': [np.asarray(1), np.asarray(2)]},
      'hello world',
      None,
  ])
  def test_pack(self, value):
    buffers = portal.pack(value)
    buffer = b''.join(buffers)
    restored = portal.unpack(buffer)
    assert portal.tree_equals(value, restored)

  def test_sharray(self):
    content = np.arange(6, dtype=np.float32).reshape(3, 2)
    value = portal.SharedArray((3, 2), np.float32)
    value.array[:] = content
    buffers = portal.pack({'foo': value})
    buffer = b''.join(buffers)
    restored = portal.unpack(buffer)
    assert restored['foo'].array.shape == (3, 2)
    assert (restored['foo'].array == content).all()
    assert restored['foo'].name == value.name
    value.close()



import threading
import time

import pytest
import portal


SERVERS = [
    portal.Server,
    portal.BatchServer,
]


class TestClient:

  def test_none_result(self):
    port = portal.free_port()
    server = portal.Server(port)
    def fn():
      pass
    server.bind('fn', fn)
    server.start(block=False)
    client = portal.Client('localhost', port)
    assert client.fn().result() is None
    client.close()
    server.close()

  def test_manual_connect(self):
    port = portal.free_port()
    client = portal.Client('localhost', port, autoconn=False)
    assert not client.connected
    server = portal.Server(port)
    server.bind('fn', lambda x: x)
    server.start(block=False)
    client.connect()
    assert client.connected
    assert client.fn(12).result() == 12
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(10))
  def test_manual_reconnect(self, repeat):
    port = portal.free_port()
    server = portal.Server(port)
    server.bind('fn', lambda x: x)
    server.start(block=False)
    client = portal.Client('localhost', port, autoconn=False)
    client.connect()
    assert client.fn(1).result() == 1
    server.close()

    assert len(client.futures) == 0
    assert len(client.errors) == 0
    try:
      future = client.fn(2)
      try:
        future.result()
        assert False
      except portal.Disconnected:
        assert True
    except portal.Disconnected:
      time.sleep(1)
      future = None
    assert len(client.futures) == 0
    assert len(client.errors) == 0

    server = portal.Server(port)
    server.bind('fn', lambda x: x)
    server.start(block=False)
    client.connect()
    assert client.fn(3).result() == 3
    client.close()
    server.close()

  def test_connect_before_server(self):
    port = portal.free_port()
    results = []

    def client():
      client = portal.Client('localhost', port)
      results.append(client.fn(12).result())
      client.close()

    thread = portal.Thread(client, start=True)
    time.sleep(0.2)
    server = portal.Server(port)
    server.bind('fn', lambda x: x)
    server.start(block=False)
    thread.join()
    server.close()
    assert results[0] == 12

  def test_future_order(self):
    port = portal.free_port()
    server = portal.Server(port)
    server.bind('fn', lambda x: x)
    server.start(block=False)
    client = portal.Client('localhost', port)
    future1 = client.fn(1)
    future2 = client.fn(2)
    future3 = client.fn(3)
    assert future2.result() == 2
    assert future1.result() == 1
    assert future3.result() == 3
    server.close()
    client.close()

  def test_future_timeout(self):
    port = portal.free_port()
    server = portal.Server(port)
    def fn(x):
      time.sleep(0.1)
      return x
    server.bind('fn', fn)
    server.start(block=False)
    client = portal.Client('localhost', port)
    future = client.fn(42)
    with pytest.raises(TimeoutError):
      future.result(timeout=0)
    with pytest.raises(TimeoutError):
      future.result(timeout=0.01)
    with pytest.raises(TimeoutError):
      future.result(timeout=0)
    assert future.result(timeout=1) == 42
    client.close()
    server.close()

  def test_maxinflight(self):
    port = portal.free_port()
    server = portal.Server(port)
    parallel = [0]
    lock = threading.Lock()

    def fn(data):
      with lock:
        parallel[0] += 1
        assert parallel[0] <= 2
      time.sleep(0.1)
      with lock:
        parallel[0] -= 1
      return data
    server.bind('fn', fn, workers=4)
    server.start(block=False)

    client = portal.Client('localhost', port, maxinflight=2)
    futures = [client.fn(i) for i in range(16)]
    results = [x.result() for x in futures]
    assert results == list(range(16))
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(5))
  def test_future_cleanup(self, repeat):
    port = portal.free_port()
    server = portal.Server(port)
    server.bind('fn', lambda x: x)
    server.start(block=False)
    client = portal.Client('localhost', port, maxinflight=1)
    client.fn(1)
    client.fn(2)
    future3 = client.fn(3)
    assert future3.result() == 3
    del future3
    assert len(client.futures) == 0
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(3))
  def test_future_cleanup_errors(self, repeat):
    port = portal.free_port()
    server = portal.Server(port, errors=False)
    def fn(x):
      if x == 2:
        raise ValueError(x)
      return x
    server.bind('fn', fn)
    server.start(block=False)
    client = portal.Client('localhost', port, maxinflight=1)
    client.fn(1)
    client.fn(2)
    time.sleep(0.2)
    with pytest.raises(RuntimeError):
      client.fn(3)
    assert client.fn(3).result() == 3
    client.close()
    server.close()

  @pytest.mark.parametrize('repeat', range(5))
  def test_client_threadsafe(self, repeat, users=16):
    port = portal.free_port()
    server = portal.Server(port)
    server.bind('fn', lambda x: x, workers=4)
    server.start(block=False)
    client = portal.Client('localhost', port, maxinflight=8)
    barrier = threading.Barrier(users)

    def user():
      barrier.wait()
      for x in range(4):
        assert client.fn(x).result() == x

    portal.run([portal.Thread(user) for _ in range(users)])
    server.close()
    client.close()

  @pytest.mark.parametrize('repeat', range(5))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_maxinflight_disconnect(self, repeat, Server):
    port = portal.free_port()
    a = threading.Barrier(2)
    b = threading.Barrier(2)

    def server():
      def fn(x):
        if x == 1:
          a.wait()
        time.sleep(0.1)
        return x
      server = Server(port)
      server.bind('fn', fn, workers=2)
      server.start(block=False)
      a.wait()
      server.close()
      server = Server(port)
      server.bind('fn', fn)
      server.start(block=False)
      b.wait()
      server.close()

    def client():
      client = portal.Client('localhost', port, maxinflight=2)
      futures = [client.fn(x) for x in range(5)]
      results = [x.result() for x in futures]
      assert results == list(range(5))
      b.wait()
      client.close()

    portal.run([
      portal.Thread(server),
      portal.Thread(client),
    ])

  @pytest.mark.parametrize('repeat', range(10))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_server_drops_autoconn(self, repeat, Server):
    port = portal.free_port()
    a = threading.Barrier(2)
    b = threading.Barrier(2)
    c = threading.Barrier(2)

    def server():
      server = Server(port)
      server.bind('fn', lambda x: x)
      server.start(block=False)
      a.wait()
      server.close()
      stats = server.stats()
      assert stats['numrecv'] == 1
      assert stats['numsend'] == stats['numrecv']
      b.wait()
      server = Server(port)
      server.bind('fn', lambda x: x)
      server.start(block=False)
      c.wait()
      server.close()

    def client():
      client = portal.Client(
          'localhost', port, maxinflight=1, autoconn=True)
      assert client.fn(1).result() == 1
      a.wait()
      b.wait()
      assert client.fn(2).result() == 2
      time.sleep(0.1)
      assert client.fn(3).result() == 3
      c.wait()
      client.close()

    portal.run([
        portal.Thread(server),
        portal.Thread(client),
    ])

  @pytest.mark.parametrize('repeat', range(3))
  @pytest.mark.parametrize('Server', SERVERS)
  def test_server_drops_manual(self, repeat, Server):
    port = portal.free_port()
    a = threading.Barrier(2)
    b = threading.Barrier(2)

    def server():
      server = Server(port)
      server.bind('fn', lambda x: x)
      server.start(block=False)
      a.wait()
      server.close()
      stats = server.stats()
      assert stats['numrecv'] == 1
      assert stats['numsend'] == stats['numrecv']
      server = Server(port)
      server.bind('fn', lambda x: x)
      server.start(block=False)
      b.wait()
      server.close()

    def client():
      client = portal.Client(
          'localhost', port, maxinflight=1, autoconn=False)
      client.connect()
      assert client.fn(1).result() == 1
      a.wait()
      time.sleep(0.1)
      with pytest.raises(portal.Disconnected):
        client.fn(3).result()
      client.connect()
      assert client.fn(3).result() == 3
      b.wait()
      client.close()

    portal.run([
        portal.Thread(server),
        portal.Thread(client),
    ])



import ctypes
import errno
import os
import socket
import sys
import threading
import time

import psutil


def run(workers, duration=None):
  [None if x.started else x.start() for x in workers]
  start = time.time()
  while True:
    time.sleep(0.1)
    if duration and time.time() - start >= duration:
      print(f'Shutting down workers after {duration} seconds.')
      [x.kill() for x in workers]
      return
    if all(x.exitcode == 0 for x in workers):
      print('All workers terminated successfully.')
      return
    errored = [x for x in workers if x.exitcode not in (None, 0)]
    if errored:
      time.sleep(0.1)  # Wait for workers to print their error messages.
      name = errored[0].name
      code = errored[0].exitcode
      print(f"Shutting down workers due to crash in '{name}' ({code}).")
      [x.kill() for x in workers]
      raise RuntimeError(f"'{name}' crashed with exit code {code}")


def kill_thread(threads, timeout=1):
  threads = threads if isinstance(threads, (list, tuple)) else [threads]
  for thread in threads:
    if thread.native_id is None:
      # Wait because thread may currently be starting.
      time.sleep(0.2)
    matches = [k for k, v in threading._active.items() if v is thread]
    if not matches:
      continue
    ident = matches[0]
    result = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(ident), ctypes.py_object(SystemExit))
    if result > 1:
      ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(ident), None)
  start = time.time()
  [x.join(max(0.1, timeout - (time.time() - start))) for x in threads]
  for thread in threads:
    if thread.is_alive():
      print('Killed thread is still alive.')


def kill_proc(procs, timeout=1):
  def eachproc(fn, procs):
    result = []
    for proc in list(procs):
      try:
        result.append(fn(proc))
      except psutil.NoSuchProcess:
        pass
    return result
  # Collect all sub processes.
  procs = procs if isinstance(procs, (list, tuple)) else [procs]
  procs = eachproc(
      lambda p: psutil.Process(p) if isinstance(p, int) else p, procs)
  eachproc(lambda p: procs.extend(p.children(recursive=True)), procs)
  procs = list(set(procs))
  # Send SIGINT to attempt graceful shutdown.
  eachproc(lambda p: p.terminate(), procs)
  _, procs = psutil.wait_procs(procs, timeout)
  # Send SIGTERM to remaining processes to force exit.
  eachproc(lambda p: p.kill(), procs)
  # Should never happen but print warning if any survived.
  eachproc(lambda p: (
      print('Killed subprocess is still alive.')
      if proc_alive(p.pid) else None), procs)


def proc_alive(pid):
  try:
    if psutil.Process(pid).status() == psutil.STATUS_ZOMBIE:
      return False
  except psutil.NoSuchProcess:
    return False
  try:
    os.kill(pid, 0)
  except OSError as e:
    if e.errno == errno.ESRCH:
      return False
  return True


def free_port():
  # Return a port that is currently free. This function is not thread or
  # process safe, because there is no way to guarantee that the port will still
  # be free at the time it will be used.
  sock = socket.socket()
  sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
  sock.bind(('localhost', 0))
  port = sock.getsockname()[1]
  sock.close()
  return port


def style(color=None, background=None, bold=None, underline=None, reset=None):
  if not sys.stdout.isatty():
    return ''
  escseq = lambda parts: '\033[' + ';'.join(parts) + 'm'
  colors = dict(
      black=0, red=1, green=2, yellow=3, blue=4, magenta=5, cyan=6, white=7)
  parts = []
  if reset:
    parts.append(escseq('0'))
  if color or bold or underline:
    args = ['3' + (str(colors[color]) if color else '9')]
    bold and args.append('1')
    underline and args.append('4')
    parts.append(escseq(args))
  if background:
    parts.append(escseq('4' + str(colors[background])))
  return ''.join(parts)



import threading
import time

import pytest
import portal


class TestThread:

  def test_exitcode(self):
    worker = portal.Thread(lambda: None, start=True)
    worker.join()
    assert worker.exitcode == 0
    worker = portal.Thread(lambda: 42, start=True)
    worker.join()
    assert worker.exitcode == 42

  def test_kill(self):
    def fn():
      while True:
        time.sleep(0.1)
    worker = portal.Thread(fn, start=True)
    worker.kill()
    assert not worker.running
    assert worker.exitcode == 2

  def test_error(self):
    def container():
      def fn():
        raise KeyError('foo')
      portal.Thread(fn, start=True).join()
    worker = portal.Process(container, start=True)
    worker.join()
    assert not worker.running
    assert worker.exitcode == 1

  def test_error_with_children(self):
    def container():
      def hang():
        while True:
          time.sleep(0.1)
      children = []
      def fn():
        children.append(portal.Process(hang, start=True))
        children.append(portal.Thread(hang, start=True))
        time.sleep(0.1)
        raise KeyError('foo')
      portal.Thread(fn, start=True).join()
    worker = portal.Process(container, start=True)
    worker.join()
    assert not worker.running
    assert worker.exitcode == 1

  @pytest.mark.parametrize('repeat', range(5))
  def test_kill_with_subthread(self, repeat):
    barrier = threading.Barrier(2)
    child = [None]
    def outer():
      child[0] = portal.Thread(inner)
      child[0].start()
      while True:
        time.sleep(0.1)
    def inner():
      barrier.wait()
      while True:
        time.sleep(0.1)
    worker = portal.Thread(outer, start=True)
    barrier.wait()
    worker.kill()
    time.sleep(0.1)
    assert not worker.running
    assert not child[0].running
    assert worker.exitcode == 2
    assert child[0].exitcode == 2

  @pytest.mark.parametrize('repeat', range(5))
  def test_kill_with_subproc(self, repeat):
    ready = portal.context.mp.Event()
    proc = [None]
    def outer():
      proc[0] = portal.Process(inner, ready, start=True)
      while True:
        time.sleep(0.1)
    def inner(ready):
      ready.set()
      while True:
        time.sleep(0.1)
    worker = portal.Thread(outer, start=True)
    ready.wait()
    worker.kill()
    assert not worker.running
    assert not proc[0].running
    assert worker.exitcode == 2
    assert proc[0].exitcode < 0
